
##
## Reference implementation of the (1+1) Limited Memory Matrix Adaptation
## Evolution Strategy; (1+1)-LM-MA-ES.
##
## Copyright (C) 2024 Tobias Glasmachers
##
## Permission is hereby granted, free of charge, to any person obtaining a
## copy of this software and associated documentation files (the
## "Software"), to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:
##
## The above copyright notice and this permission notice shall be included
## in all copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
## OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
## IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
## CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
## TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
## SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##


import numpy as np
import sys
import time


def lm_ma_es_1p1(x, sigma, f, budget, verbosity = 1):
	start = time.time()

	# preparation
	n = len(x)                                 # search space dimensionality
	k = int(4 + np.floor(3 * np.log(n)))       # number of directions in the covariance model
	alpha = 2 / n                              # step size adaptation learning rate
	c_del = np.zeros((k,))                     # information deletion learning rate
	c_new = np.zeros((k,))                     # information update learning rate
	for i in range(k):
		c_del[i] = 1 / (4.0**i * n)
		c_new[i] = np.sqrt(2) / n * 6**(-i/2)
	c_keep = (1 - c_del).reshape(k, 1)         # information to keep
	m = np.zeros((k, n))                       # k path vectors (covariance information)

	if verbosity >= 1:
		print("[1+1 LM-MA-ES]  f=" + f.name + "  n=" + str(n) + "  k=" + str(k) + "  budget=" + str(budget))

	prev_report = 0
	def report(fe, fitness, sigma, force):
		nonlocal prev_report
		t = time.time()
		if force or t - prev_report >= 0.2:
			sys.stdout.write('\b' * 100);
			sys.stdout.write("    #FE: " + str(fe) + "  fitness: " + str(fitness) + "  step size: " + str(sigma) + "     \b\b\b\b\b")
			sys.stdout.flush()
			prev_report = t

	# evaluate parent
	fx = f(x)

	# ring buffer
	buffer = [1e100] * 256
	buffer[0] = fx
	bufferindex = 1

	# generation loop
	t = 1   # one individual is already evaluated
	while t < budget:
		# generation counter
		t += 1

		# sample offspring
		z = np.random.randn(n)
		d = z + (m @ z) @ m

		# evaluate offspring
		y = x + sigma * d
		fy = f(y)

		# check for success
		success = (fy <= fx)

		if success:
			# update mean
			x = y
			fx = fy

			# update direction vectors
			m *= c_keep
			m += np.outer(c_new, z)

		# update step size
		sigma *= np.exp((1 if success else -0.25) * alpha)

		# detect convergence or stall
		if n * sigma < 1e-10:
			report(t, fx, sigma, True)
			print("  step size converged (" + str(time.time() - start) + " seconds)")
			return x, t
		elif n * sigma > 1e5:
			report(t, fx, sigma, True)
			print("  step size diverged (" + str(time.time() - start) + " seconds)")
			return x, t
		elif buffer[bufferindex] - fx == 0:
			report(t, fx, sigma, True)
			print("  fitness converged (" + str(time.time() - start) + " seconds)")
			return x, t
		buffer[bufferindex] = fx
		bufferindex = (bufferindex + 1) % len(buffer)

		# status report
		report(t, fx, sigma, False)

	# we ran out of budget
	report(budget, fx, sigma, True)
	print("  budget exceeded (" + str(time.time() - start) + " seconds)")
	return x, budget
